unit UFormCheck;

{$mode delphi}

interface

uses
  Classes, SysUtils, FileUtil, Forms, Controls, Graphics, Dialogs, StdCtrls,
  ExtCtrls, ComCtrls, UAcronym, URegistry, Registry, UCommon;

type

  { TFormCheck }

  TFormCheck = class(TForm)
    ButtonLoadFromFile: TButton;
    ButtonAcronymsContent: TButton;
    ButtonAcronymsSummary: TButton;
    ButtonCheck: TButton;
    EditSummaryStart: TEdit;
    EditSummaryEnd: TEdit;
    GroupBox1: TGroupBox;
    GroupBox2: TGroupBox;
    Label1: TLabel;
    Label2: TLabel;
    LabelAcronymCountContent: TLabel;
    LabelAcronymCountSummary: TLabel;
    MemoDocument: TMemo;
    MemoReport: TMemo;
    OpenDialog1: TOpenDialog;
    PageControl1: TPageControl;
    Panel1: TPanel;
    Panel2: TPanel;
    Splitter1: TSplitter;
    TabSheetSource: TTabSheet;
    TabSheetReport: TTabSheet;
    procedure ButtonAcronymsSummaryClick(Sender: TObject);
    procedure ButtonAcronymsContentClick(Sender: TObject);
    procedure ButtonCheckClick(Sender: TObject);
    procedure ButtonLoadFromFileClick(Sender: TObject);
    procedure FormClose(Sender: TObject; var CloseAction: TCloseAction);
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure FormShow(Sender: TObject);
  private
    AcronymDbSummary: TAcronymDb;
    AcronymDbContent: TAcronymDb;
    LastDocumentFileName: string;
    function SearchLine(Lines: TStrings; Text: string; Start: Integer = 0): Integer;
    function SearchLineReverse(Lines: TStrings; Text: string; Start: Integer = -1): Integer;
    procedure FindInSummary;
    procedure FindInContent;
    function ParseMeaning(Acronym, Text: string; StartIndex: Integer;
      out Meaning: string; DashSeparator: Boolean = False): Boolean;
    function IsUppercaseAlpha(Text: string): Boolean;
    function IsAlpha(Text: string): Boolean;
    function IsAcronym(Text: string): Boolean;
    procedure ReportDifferencies;
    function WordContainsLetters(Text, Letters: string): Boolean;
  public
    procedure UpdateInterface;
    procedure LoadConfig;
    procedure SaveConfig;
  end;

var
  FormCheck: TFormCheck;


implementation

{$R *.lfm}

uses
  UFormAcronyms, UCore;

resourcestring
  SAcronymCountContent = 'Content acronym count:';
  SAcronymCountSummary = 'Summary acronym count:';
  SDuplicateAcronymContent = 'Warning: Duplicate acronym %s with "%s" in document body.';
  SDuplicateAcronymSummary = 'Warning: Duplicate acronym %s with "%s" in acronym summary.';
  SMissingAcronymContent = 'Warning: Document body acronym %s with meaning "%s" missing from acronym summary.';
  SMissingAcronymSummary = 'Warning: Summary acronym %s with meaning "%s" missing from document body.';
  SPluralAcronym = 'Note: Acronym %s is defined as plural in document body.';

{ TFormCheck }

procedure TFormCheck.ButtonCheckClick(Sender: TObject);
begin
  MemoReport.Lines.Clear;
  FindInSummary;
  FindInContent;
  ReportDifferencies;
  UpdateInterface;
  TabSheetReport.Show;
end;

procedure TFormCheck.ButtonLoadFromFileClick(Sender: TObject);
begin
  OpenDialog1.InitialDir := ExtractFileDir(LastDocumentFileName);
  OpenDialog1.FileName := ExtractFileName(LastDocumentFileName);
  if OpenDialog1.Execute then begin
    LastDocumentFileName := OpenDialog1.FileName;
    MemoDocument.Lines.LoadFromFile(OpenDialog1.FileName);
    TabSheetSource.Show;
  end;
end;

procedure TFormCheck.FormClose(Sender: TObject; var CloseAction: TCloseAction);
begin
  Core.PersistentForm1.Save(Self);
end;

procedure TFormCheck.FormCreate(Sender: TObject);
begin
  AcronymDbSummary := TAcronymDb.Create;
  AcronymDbContent := TAcronymDb.Create;
  Core.CoolTranslator1.TranslateComponentRecursive(Self);
  Core.ThemeManager.UseTheme(Self);
end;

procedure TFormCheck.FormDestroy(Sender: TObject);
begin
  AcronymDbSummary.Free;
  AcronymDbContent.Free;
end;

procedure TFormCheck.FormShow(Sender: TObject);
begin
  Core.PersistentForm1.Load(Self);
  if FileExists(LastDocumentFileName) then MemoDocument.Lines.LoadFromFile(LastDocumentFileName);
  UpdateInterface;
end;

procedure TFormCheck.ButtonAcronymsContentClick(Sender: TObject);
var
  FormAcronyms: TFormAcronyms;
begin
  FormAcronyms := TFormAcronyms.Create(Self);
  try
    FormAcronyms.Acronyms := AcronymDbContent.Acronyms;
    FormAcronyms.ShowModal;
  finally
    FreeAndNil(FormAcronyms);
  end;
end;

procedure TFormCheck.ButtonAcronymsSummaryClick(Sender: TObject);
var
  FormAcronyms: TFormAcronyms;
begin
  FormAcronyms := TFormAcronyms.Create(Self);
  try
    FormAcronyms.Acronyms := AcronymDbSummary.Acronyms;
    FormAcronyms.ShowModal;
  finally
    FreeAndNil(FormAcronyms);
  end;
end;

function TFormCheck.SearchLine(Lines: TStrings; Text: string; Start: Integer = 0): Integer;
begin
  Result := Start;
  while (Result < Lines.Count) and (Pos(Text, Lines[Result]) = 0) do Inc(Result);
  if Result >= Lines.Count then Result := -1;
end;

function TFormCheck.SearchLineReverse(Lines: TStrings; Text: string;
  Start: Integer = -1): Integer;
begin
  if Start = -1 then Result := Lines.Count - 1
    else Result := Start;
  while (Result >= 0) and (Pos(Text, Lines[Result]) = 0) do Dec(Result);
end;

procedure TFormCheck.FindInSummary;
var
  AcronymSectionStart: Integer;
  AcronymSectionEnd: Integer;
  I: Integer;
  Line: string;
  Acronym: string;
  Meaning: string;
begin
  AcronymDbSummary.Acronyms.Clear;

  AcronymSectionStart := SearchLineReverse(MemoDocument.Lines, EditSummaryStart.Text);
  if AcronymSectionStart <> -1 then begin
    AcronymSectionEnd := SearchLine(MemoDocument.Lines, EditSummaryEnd.Text, AcronymSectionStart + 1);
    if AcronymSectionEnd <> -1 then begin
      for I := AcronymSectionStart + 1 to AcronymSectionEnd - 1 do begin
        Line := Trim(MemoDocument.Lines[I]);
        Line := StringReplace(Line, #9, ' ', [rfReplaceAll]);
        if Line <> '' then begin
          if Pos(' ', Line) > 0 then begin
            Acronym := Copy(Line, 1, Pos(' ', Line) - 1);
            if IsUppercaseAlpha(Acronym) then begin
              Meaning := Trim(Copy(Line, Pos(' ', Line) + 1, Length(Line)));
              if Assigned(AcronymDbSummary.SearchAcronym(Acronym, Meaning)) then
                MemoReport.Lines.Add(Format(SDuplicateAcronymSummary, [Acronym, Meaning]))
                else AcronymDbSummary.AddAcronym(Acronym, Meaning);
            end;
          end;
        end;
      end;
    end;
  end;
end;

procedure TFormCheck.FindInContent;
var
  Text: string;
  StartIndex: Integer;
  EndIndex: Integer;
  Acronym: string;
  Meaning: string;
  Meaning1: string;
  Meaning2: string;
  HasMeaning1: Boolean;
  HasMeaning2: Boolean;
  Plural: Boolean;
begin
  AcronymDbContent.Acronyms.Clear;

  // Find acronyms usage in text
  Text := MemoDocument.Lines.Text;
  Text := StringReplace(Text, #9, ' ', [rfReplaceAll]);
  Text := StringReplace(Text, LineEnding, ' ', [rfReplaceAll]);
  StartIndex := 1;
  repeat
    StartIndex := PosFromIndex('(', Text, StartIndex);
    if StartIndex <> 0 then begin
      EndIndex := PosFromIndex(')', Text, StartIndex);
      if EndIndex <> 0 then begin
        Acronym := Trim(Copy(Text, StartIndex + 1, EndIndex - StartIndex - 1));
        // Allow plural acronyms with ending 's' character
        if (Length(Acronym) >= 1) and (Acronym[Length(Acronym)] = 's') then begin
          Acronym := Copy(Acronym, 1, Length(Acronym) - 1);
          Plural := True;
        end else Plural := False;
        if IsAcronym(Acronym) then begin
          HasMeaning1 := ParseMeaning(Acronym, Text, StartIndex - 1, Meaning1);
          if HasMeaning1 then Meaning := Meaning1;
          HasMeaning2 := ParseMeaning(Acronym, Text, StartIndex - 1, Meaning2, True);
          if HasMeaning2 then Meaning := Meaning2;
          if HasMeaning1 and HasMeaning2 then begin
            if Length(Meaning1) > Length(Meaning2) then Meaning := Meaning1
              else Meaning := Meaning2;
          end;
          if HasMeaning1 or HasMeaning2 then begin
            // If plural acronym then try to remove ending 's' character from the meaning
            if Plural then MemoReport.Lines.Add(Format(SPluralAcronym, [Acronym]));
            if Plural and (Length(Meaning) >= 1) and (Meaning[Length(Meaning)] = 's') then begin
              Meaning := Copy(Meaning, 1, Length(Meaning) - 1);
            end;
            if Assigned(AcronymDbContent.SearchAcronym(Acronym, Meaning)) then
              MemoReport.Lines.Add(Format(SDuplicateAcronymContent, [Acronym, Meaning]))
              else AcronymDbContent.AddAcronym(Acronym, Meaning);
          end;
        end;
      end;
      Inc(StartIndex);
    end;
  until StartIndex = 0;
end;

function TFormCheck.ParseMeaning(Acronym, Text: string; StartIndex: Integer;
  out Meaning: string; DashSeparator: Boolean): Boolean;
var
  StartIndex2: Integer;
  StartIndex3: Integer;
  StartIndex4: Integer;
  LetterIndex: Integer;
  OneWord: string;
  WordLetterIndex: Integer;
  WordCount: Integer;
  WordCountWrong: Integer;
begin
  Result := True;
  Meaning := '';
  StartIndex2 := StartIndex;
  LetterIndex := Length(Acronym);
  WordCount := 0;
  WordCountWrong := 0;
  while Length(Acronym) > 0 do begin
    StartIndex3 := PosFromIndexReverse(' ', Text, StartIndex2);
    if DashSeparator then begin
      StartIndex4 := PosFromIndexReverse('-', Text, StartIndex2);
      if StartIndex4 > StartIndex3 then StartIndex3 := StartIndex4;
    end;

    if StartIndex3 = 0 then Break;
    OneWord := Copy(Text, StartIndex3 + 1, StartIndex2 - StartIndex3);
    if OneWord = '$' then begin
      // Avoid parsing Bash variables
      Result := False;
      Exit;
    end;
    if Trim(OneWord) = '' then begin
      StartIndex2 := StartIndex3 - 1;
      Continue;
    end;
    // Is first letter capital?
    if (Length(OneWord) > 0) and IsAlpha(OneWord[1]) then begin
      WordLetterIndex := PosFromIndexReverse(LowerCase(OneWord[1]), LowerCase(Copy(Acronym, 1, LetterIndex)), LetterIndex);
      if WordLetterIndex > 0 then begin
        // First letter was found in acronym
        if WordLetterIndex <= LetterIndex then begin
          if not WordContainsLetters(LowerCase(OneWord), LowerCase(Copy(Acronym, WordLetterIndex, LetterIndex - WordLetterIndex + 1))) then begin
            Result := False;
            Exit;
          end;
          LetterIndex := WordLetterIndex - 1;
        end else begin
          Dec(LetterIndex);
        end;
        WordCountWrong := 0;
      end else begin
        Inc(WordCountWrong);
        if WordCountWrong > 1 then begin
          Result := False;
          Exit;
        end;
      end;
    end else begin
      Inc(WordCountWrong);
      if WordCountWrong > 1 then begin
        Result := False;
        Exit;
      end;
    end;
    StartIndex2 := StartIndex3 - 1;
    if LetterIndex < 1 then Break;
    Inc(WordCount);
    if WordCount > 2 * Length(Acronym) then begin
      // False acronym in braces with too much words
      Result := False;
      Exit;
    end;
  end;
  Meaning := Trim(Copy(Text, StartIndex2 + 1, StartIndex - StartIndex2 - 1));
end;

function TFormCheck.IsUppercaseAlpha(Text: string): Boolean;
var
  I: Integer;
begin
  I := 1;
  Result := True;
  while (I <= Length(Text)) do begin
    if not (Text[I] in ['A'..'Z']) then begin
      Result := False;
      Break;
    end;
    Inc(I);
  end;
end;

function TFormCheck.IsAlpha(Text: string): Boolean;
var
  I: Integer;
begin
  I := 1;
  Result := True;
  while (I <= Length(Text)) do begin
    if not (Text[I] in ['A'..'Z']) and not (Text[I] in ['a'..'z']) then begin
      Result := False;
      Break;
    end;
    Inc(I);
  end;
end;

function TFormCheck.IsAcronym(Text: string): Boolean;
const
  MinAcronymLength = 2;
begin
  Result := (Length(Text) >= MinAcronymLength) and IsUppercaseAlpha(Text);
end;

procedure TFormCheck.ReportDifferencies;
var
  I: Integer;
  J: Integer;
  Acronym: TAcronym;
  Meaning: TAcronymMeaning;
begin
  for I := 0 to AcronymDbContent.Acronyms.Count - 1 do begin
    Acronym := TAcronym(AcronymDbContent.Acronyms[I]);
    for J := 0 to Acronym.Meanings.Count - 1 do begin
      Meaning := TAcronymMeaning(Acronym.Meanings[J]);
      if not Assigned(AcronymDbSummary.SearchAcronym(Acronym.Name, Meaning.Name, [sfCaseInsensitive])) then
        MemoReport.Lines.Add(Format(SMissingAcronymContent, [Acronym.Name, Meaning.Name]));
    end;
  end;

  for I := 0 to AcronymDbSummary.Acronyms.Count - 1 do begin
    Acronym := TAcronym(AcronymDbSummary.Acronyms[I]);
    for J := 0 to Acronym.Meanings.Count - 1 do begin
      Meaning := TAcronymMeaning(Acronym.Meanings[J]);
      if not Assigned(AcronymDbContent.SearchAcronym(Acronym.Name, Meaning.Name, [sfCaseInsensitive])) then
        MemoReport.Lines.Add(Format(SMissingAcronymSummary, [Acronym.Name, Meaning.Name]));
    end;
  end;
end;

function TFormCheck.WordContainsLetters(Text, Letters: string): Boolean;
var
  I: Integer;
  LetterIndex: Integer;
begin
  Result := True;
  for I := 1 to Length(Letters) do begin
    LetterIndex := Pos(Letters[I], Text);
    if LetterIndex > 0 then begin
      Text := Copy(Text, LetterIndex + 1, Length(Text));
    end else begin
      Result := False;
      Break;
    end;
  end;
end;

procedure TFormCheck.UpdateInterface;
begin
  LabelAcronymCountContent.Caption := SAcronymCountContent + ' ' + IntToStr(AcronymDbContent.GetMeaningsCount);
  LabelAcronymCountSummary.Caption := SAcronymCountSummary + ' ' + IntToStr(AcronymDbSummary.GetMeaningsCount);
end;

procedure TFormCheck.LoadConfig;
begin
  with TRegistryEx.Create do
  try
    RootKey := HKEY_CURRENT_USER;
    OpenKey(DefaultRegKey, True);
    EditSummaryStart.Text := ReadStringWithDefault('SummaryStart', 'ACRONYMS AND ABBREVIATIONS');
    EditSummaryEnd.Text := ReadStringWithDefault('SummaryEnd', 'Appendix');
    LastDocumentFileName := ReadStringWithDefault('LastDocumentFileName', '')
  finally
    Free;
  end;
end;

procedure TFormCheck.SaveConfig;
begin
  with TRegistryEx.Create do
  try
    RootKey := HKEY_CURRENT_USER;
    OpenKey(DefaultRegKey, True);
    WriteString('SummaryStart', EditSummaryStart.Text);
    WriteString('SummaryEnd', EditSummaryEnd.Text);
    WriteString('LastDocumentFileName', LastDocumentFileName);
  finally
    Free;
  end;
end;

end.

